/*
 * Copyright (c) 1994 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Dem/dcwdem.h>
#include <limits.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#if defined(__linux__)
#define LOCAL_BYTE_ORDER 'I'
#else
#define LOCAL_BYTE_ORDER 'M'
#endif

DcwDem::DcwDem() 
: DEM()
{
    emax = INT_MIN;
    emin = INT_MAX;
}

void DcwDem::read(char* path) {
    char* headerpath = strdup(path);
    int len = strlen(headerpath);
    if (path[len-1]=='z' && path[len-2]=='g' && path[len-3]=='.') {
	len -= 3;
	headerpath[len]='\0';
    }
    else if (path[len-1]=='Z' && path[len-2]=='.') {
	len -= 2;
	headerpath[len]='\0';
    }
    headerpath[len-1]='R';
    headerpath[len-2]='D';
    headerpath[len-3]='H';
    FILE* header;
    if ((header = fopen(headerpath, "r"))==NULL) {
	printf("couldn't open header file\n");
	return;
    }
    char line[80];
    fgets(line, 80, header);
    sscanf(line, "%*s %c", &byteorder);
    fgets(line, 80, header);
    fgets(line, 80, header);
    sscanf(line, "%*s %d", &r);
    fgets(line, 80, header);
    sscanf(line, "%*s %d", &c);
    fgets(line, 80, header);
    fgets(line, 80, header);
    fgets(line, 80, header);
    fgets(line, 80, header);
    fgets(line, 80, header);
    fgets(line, 80, header);
    sscanf(line, "%*s %lf", &nwe);
    fgets(line, 80, header);
    sscanf(line, "%*s %lf", &nwn);
    fclose(header);
    _swe = _nwe;
    _swn = _nwn - ((r * 30.0)/3600.0);
    _nee = _nwe + ((c * 30.0)/3600.0);
    _nen = _nwn;
    _see = _nee;
    _sen = _swn;
    grunits = arcseconds;
    elunits = feet;
    allocate(r, c);
    FILE* f;
    int p = 0;
    len = strlen(path);
    if (path[len-1]=='Z' && path[len-2]=='.') {
	char* cmd = new char[len+6];
	strcpy(cmd, "zcat ");
	strcat(cmd, path);
	if ((f = popen(cmd, "r")) == NULL) {
	    printf("popen failed\n");
	    return;
	}
	delete cmd;
	p = 1;
    }
    else if (path[len-1]=='z' && path[len-2]=='g' && path[len-3]=='.') {
	char* cmd = new char[len+11];
	strcpy(cmd, "gunzip -c ");
	strcat(cmd, path);
	if ((f = popen(cmd, "r")) == NULL) {
	    printf("popen failed\n");
	    return;
	}
	delete cmd;
	p = 1;
    }
    else if ((f = fopen(path, "r"))==NULL) {
	printf("couldn't open DCW DEM file\n");
	return;
    }
    short int v;
    short int w;
    int i;
    for (i = r-1; i >= 0; i--) {
	for (int j = 0; j < c; j++) {
	    fread(&w, sizeof(short), 1, f);
	    if (byteorder != LOCAL_BYTE_ORDER) {
		char* vb = (char*) &v;
		char* wb = (char*) &w;
		vb[0] = wb[1];
		vb[1] = wb[0];
	    }
	    else
		v = w;
	    if (v == -999)
		v = 0;
	    value(i, j, v);
	}
    }
    if (p)
	pclose(f);
    else
	fclose(f);
    for (i = r-1; i >= 0; i--) {
	for (int j = 0; j < c; j++) {
	    if (value(i,j)>emax)
		emax=value(i,j);
	    if (value(i,j)<emin)
		emin=value(i,j);
	}
    }
}

void DcwDem::print() {
    printf("Rows/Columns: %d %d\n", r, c);
    printf("Corners: SW: %0.1fE %0.1fN; NW: %0.1fE %0.1fN; NE: %0.1fE %0.1fN; SE: %0.1fE %0.1fN\n",
	   swe, swn, nwe, nwn, nee, nen, see, sen);
    for (int i = 0; i < cols(); i++) {
	printf("Column %d:", i);
	for (int j = 0; j < rows(); j++) {
	    printf(" %d", value(j, i));
	}
	printf("\n");
    }
}

void DcwDem::writerawpgm(char* path) {
    FILE* f;
    if ((f = fopen(path, "w")) == NULL) {
	printf("couldn't open file for writing\n");
	return;
    }
    fprintf(f, "P5\n%d %d\n", c, r);
    fprintf(f, "%d\n", 255);
    for (int i = r-1; i >= 0; i--) {
	for (int j = 0; j < c; j++) {
	    int v = int(rint(value(i, j) * (255.0/(emax-emin))));
	    fputc(v, f);
	}
    }
    fclose(f);
}

int DcwDem::elev_max() {
    return emax;
}
