/*
 * Copyright (c) 1996 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#ifndef demeditor_h
#define demeditor_h

#ifdef VTK

#include <TopoFace/fgeomobjs.h>
#include <MapUnidraw/mapeditor.h>
#include <DemUnidraw/demkit.h>
#include <InterViews/action.h>

class BoundedValue;
class UsgsDem;
class DataSetLODs;
class vtkDataSet;

class DataSetLODs {
public:
  vtkDataSet* dslod[4];
};

class DemEditor : public MapEditor {
public:
    DemEditor(OverlayComp*, DemKit* ok = DemKit::Instance());
    DemEditor(const char* file, DemKit* ok = DemKit::Instance());
    DemEditor(boolean initflag, DemKit* ok = DemKit::Instance());
    virtual ~DemEditor();
    virtual void AddCommands(ComTerp*);

  virtual StateVar* GetState(const char*);

    virtual void Update();
    virtual void InitCommands();
    UsgsDem* DEM() { return dem; }
    void DEM(UsgsDem* d) { dem = d; }
    void VehicleName(char* n) { _name = n; }
    char* VehicleName() { return _name; }
    void* VehicleSocket() { return _vsocket; }
    void VehicleSocket(void* v) { _vsocket = v; }
    void CameraHeight(float h) { _height = h; }
    float CameraHeight() { return _height; }

    void CameraUp();
    void CameraDown();
    void CameraYawLeft();
    void CameraYawRight();

  BoundedValue* CameraX() { return _CameraX; }
  BoundedValue* CameraY() { return _CameraY; }
  BoundedValue* CameraZ() { return _CameraZ; }
  BoundedValue* FocalPtX() { return _FocalPtX; }
  BoundedValue* FocalPtY() { return _FocalPtY; }
  BoundedValue* FocalPtZ() { return _FocalPtZ; }
  BoundedValue* ClipFront() { return _ClipFront; }
  BoundedValue* ClipBack() { return _ClipBack; }
  BoundedValue* ViewUpX() { return _ViewUpX; }
  BoundedValue* ViewUpY() { return _ViewUpY; }
  BoundedValue* ViewUpZ() { return _ViewUpZ; }
  BoundedValue* ViewAngle() { return _ViewAngle; }
  BoundedValue* Distance() { return _Distance; }

  void ShowCameraStateWindow() { ((DemKit*)_overlay_kit)->ShowCameraStateWindow(); }
  void HideCameraStateWindow() { ((DemKit*)_overlay_kit)->HideCameraStateWindow(); }
  void LoadActorsByFootprint(float*, float*);
  static vtkDataSet* GetDataSet(int i, int j, int lod);
  static FBoxObj GetBox(int i, int j);
  static void GetDataSetXY(float px, float py, int*& x, int*& y, int& n);
  char* GetFilename() { return _filename; }
  int GetCount() { return count; }
  static int GetCounter() { return counter; }
protected:
    void Init(OverlayComp* = nil, const char* = "DemEditor");

    Interactor* Interior();

  void InitLOD(char* dirname, char* docname);
  static char basename[256];
  static DataSetLODs dstable[80*80];
  static FBoxObj dsindex0[80*80];
  static FBoxObj dsindex1[40*40];
  static FBoxObj dsindex2[20*20];
  static FBoxObj dsindex3[10*10];
  static FBoxObj dsindex4[5*5];
  static int counter;

  int count;
    UsgsDem* dem;

    char* _name;
    char* _filename;

    void* _vsocket;

    float _height;

  void InitCameraValues();
    BoundedValue* _CameraX;
    BoundedValue* _CameraY;
    BoundedValue* _CameraZ;
    BoundedValue* _FocalPtX;
    BoundedValue* _FocalPtY;
    BoundedValue* _FocalPtZ;
    BoundedValue* _ClipFront;
    BoundedValue* _ClipBack;
    BoundedValue* _ViewUpX;
    BoundedValue* _ViewUpY;
    BoundedValue* _ViewUpZ;
    BoundedValue* _ViewAngle;
    BoundedValue* _Distance;

friend DemKit;
};

declareActionCallback(DemEditor)

#endif

#endif
