/*
 * Copyright (c) 1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <UtmUnidraw/utmclasses.h>
#include <UtmUnidraw/utmcmds.h>
#include <UtmUnidraw/utmeditor.h>

#include <MapUnidraw/mapcomps.h>
#include <MapUnidraw/mapimport.h>
#include <MapUnidraw/maproute.h>
#include <MapUnidraw/mapstates.h>
#include <MapUnidraw/mapviewstate.h>
#include <MapUnidraw/mapviewer.h>

#include <OverlayUnidraw/ovarrow.h>
#include <OverlayUnidraw/ovimport.h>
#include <OverlayUnidraw/ovrect.h>

#include <Map/mapfclass.h>
#include <Map/mapfeature.h>
#include <Map/mapprim.h>

#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>

#include <UniIdraw/idarrows.h>

#include <Unidraw/catalog.h>
#include <Unidraw/clipboard.h>
#include <Unidraw/manips.h>
#include <Unidraw/selection.h>
#include <Unidraw/unidraw.h>
#include <Unidraw/viewer.h>

#include <Unidraw/Commands/edit.h>

#include <Unidraw/Graphic/damage.h>
#include <Unidraw/Graphic/graphic.h>
#include <Unidraw/Graphic/polygons.h>

#include <InterViews/cursor.h>
#include <InterViews/transformer.h>
#include <InterViews/window.h>
#include <IV-2_6/InterViews/rubline.h>
#include <IV-2_6/InterViews/rubrect.h>
#include <OS/math.h>

#include <math.h>
#include <stdio.h>
#include <strstream.h>

#ifdef HAVE_ACE
#include <ace/SOCK_Stream.h>
#include <ace/SOCK_Connector.h>
#include <ace/INET_Addr.h>
#endif

static const unit = 15;

/*****************************************************************************/

ClassId UtmShowCmd::GetClassId () { return UTMSHOW_CMD; }

boolean UtmShowCmd::IsA (ClassId id) {
    return UTMSHOW_CMD == id || Command::IsA(id);
}

UtmShowCmd::UtmShowCmd (ControlInfo* c) : Command(c) { }
UtmShowCmd::UtmShowCmd (Editor* ed) : Command(ed) { }

Command* UtmShowCmd::Copy () {
    Command* copy = new UtmShowCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void UtmShowCmd::Execute () {
#if 0
    UtmEditor* editor = (UtmEditor*)GetEditor();
    editor->GetWindow()->cursor(hourglass);
    GraphicComps* top = (GraphicComps*)GetEditor()->GetComponent();
    Clipboard* clb = new Clipboard();
    
    MapViewState* viewstate = (MapViewState*)editor->GetMapViewState();
    MapFeatureClass* mapfclass = new MapFeatureClass(nil, "utmfeature");
    if (!mapfclass) return;

    int nfeats = mapfclass->nfeats();
    for (int i = 0; i<nfeats; i++) {
	MapFeature* feature = mapfclass->feature(i);
	if (feature->face_primitive()->npts()<0) continue;
	if (!viewstate->themegs(feature)) continue;
	MapFeatureComp* mfc = new MapFeatureComp(feature);
	if (mfc) 
	    clb->Append(mfc);
    }
    
    PasteCmd* pastecmd = new PasteCmd(editor, clb);
    pastecmd->Execute();
    Selection* s = GetEditor()->GetViewer()->GetSelection();
    s->Clear();
#endif
}

boolean UtmShowCmd::Reversible () { return true; }


/*****************************************************************************/

#ifdef HAVE_ACE

ClassId UtmImportByServCmd::GetClassId () { return UTMIMPORTBYSERV_CMD; }

boolean UtmImportByServCmd::IsA (ClassId id) {
    return UTMIMPORTBYSERV_CMD == id || Command::IsA(id);
}

UtmImportByServCmd::UtmImportByServCmd (ControlInfo* c) : Command(c) { 
  _impcmd = nil;
}
UtmImportByServCmd::UtmImportByServCmd (Editor* ed) : Command(ed) { 
  _impcmd = nil;
}

Command* UtmImportByServCmd::Copy () {
    Command* copy = new UtmImportByServCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void UtmImportByServCmd::ServHost (const char* host) {
  _host = (char*)host;
}

void UtmImportByServCmd::ServPort (unsigned short port) {
  _port = port;
}

void UtmImportByServCmd::ServCom (char* com) {
  _com = com;
}

void UtmImportByServCmd::ResultImportCmd (OvImportCmd* impcmd) {
  _impcmd = impcmd;
}

void UtmImportByServCmd::Execute () {
  GetEditor()->GetWindow()->cursor(hourglass);

  ACE_SOCK_Stream* _socket = new ACE_SOCK_Stream;
  ACE_SOCK_Connector* connector = new ACE_SOCK_Connector;
  ACE_INET_Addr* addr = new ACE_INET_Addr(_port, _host);

  if (connector->connect (*_socket, *addr) == -1) {
    GetEditor()->GetWindow()->cursor(arrow);
    ACE_ERROR ((LM_ERROR, "%p\n", "open"));
    return;
  }
  if (_socket->send (_com, strlen(_com)) == -1) {
    GetEditor()->GetWindow()->cursor(arrow);
    ACE_ERROR ((LM_ERROR, "%p\n", "send"));
    return;
  }
  if (_impcmd) {
    filebuf* _filebuf = new filebuf();
    boolean valid = _filebuf->attach(_socket->get_handle()) != 0;
    istream*_inptr = new istream(_filebuf);
    ((OverlayIdrawComp*)GetEditor()->GetComponent())->ResetIndexedGS();
    _impcmd->instream(_inptr);
    _impcmd->Execute();
  }
  if (_socket->send ("quit\n", strlen("quit\n")) == -1) {
    GetEditor()->GetWindow()->cursor(arrow);
    ACE_ERROR ((LM_ERROR, "%p\n", "send"));
    return;
  }
}

boolean UtmImportByServCmd::Reversible () { return false; }

/****************************************************************************/

UtmMapImportByServCmd::UtmMapImportByServCmd (ControlInfo* c) : UtmImportByServCmd(c) { }
UtmMapImportByServCmd::UtmMapImportByServCmd (Editor* ed) : UtmImportByServCmd(ed) { }

Command* UtmMapImportByServCmd::Copy () {
    Command* copy = new UtmMapImportByServCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void UtmMapImportByServCmd::Execute () {
  Catalog* catalog = unidraw->GetCatalog();
  const char* portstr = catalog->GetAttribute("mapservport");
  u_short MAPSERV_PORT = atoi(portstr);
  const char *const MAPSERV_HOST = catalog->GetAttribute("mapservhost");
  const char *mapserv_host = MAPSERV_HOST;
  u_short mapserv_port  = MAPSERV_PORT;
  ServHost(mapserv_host);
  ServPort(mapserv_port);
  ServCom("print(loadfc(\"/data/maps/vpf/killeen\" \"us064462\" \"trn\" \"trdline\") :paint)\n");
  UtmEditor* editor = (UtmEditor*)GetEditor();
  MapImportCmd* _import_cmd = new MapImportCmd(editor);
  ResultImportCmd(_import_cmd);
  UtmImportByServCmd::Execute();
}

#endif

/****************************************************************************/

ClassId UtmPosTool::GetClassId () { return UTMPOS_TOOL; }

boolean UtmPosTool::IsA (ClassId id) {
    return UTMPOS_TOOL == id || GraphicCompTool::IsA(id);
}

Tool* UtmPosTool::Copy () { return new UtmPosTool(CopyControlInfo()); }

UtmPosTool::UtmPosTool(ControlInfo* info)
:GraphicCompTool()
{
    if (info)
	SetControlInfo(info);
    ArrowLine* ml = new ArrowLine(
        0, 0, unit, unit, true, false, 2., stdgraphic
    );
    Catalog* catalog = unidraw->GetCatalog();
    PSBrush* br = catalog->FindBrush(0xffff, 2);
    ml->SetBrush(br);
    PSColor* col = catalog->FindColor("Red");
    ml->SetColors(col, col);
    ArrowLineOvComp* mlComp = new ArrowLineOvComp(ml);
    Init(mlComp);
}

Command* UtmPosTool::InterpretManipulator (Manipulator* m) {
    DragManip* dm = (DragManip*) m;
    MapEditor* ed = (MapEditor*) dm->GetViewer()->GetEditor();
    Transformer* rel = dm->GetTransformer();
    MacroCmd* cmd = new MacroCmd(ed);
    Transformer* trans = ed->GetMapViewState()->deftrans();

    RubberLine* rl = (RubberLine*) dm->GetRubberband();

    Coord px0, py0, px1, py1;
    int pt;
    rl->GetCurrent(px0, py0, px1, py1);

    if (rel != nil) {
      rel = new Transformer(rel);
      rel->Invert();
    }

    if (px0 != px1 || py0 != py1) {
        float fpx, fpy, ffx, ffy;
	rel->transform(px0, py0, fpx, fpy);
	rel->transform(px1, py1, ffx, ffy);
	printf("rl: %d %d %d %d; trl: %f %f %f %f\n",
	       px0, py0, px1, py1, fpx, fpy, ffx, ffy);
	if (trans) {
	  float tpx, tpy, tfx, tfy;
	  trans->inverse_transform(fpx, fpy, tpx, tpy);
	  trans->inverse_transform(ffx, ffy, tfx, tfy);
	  printf("UTM: P1: %fE %fN; P2: %fE %fN\n", tpx, tpy, tfx, tfy);
	}

	if (ed->GetCameraComp()) {
	  Viewer* viewer = ed->GetViewer();    
	  OverlayView* view =
	    ((OverlayComp*)ed->GetCameraComp())->FindView(ed->GetViewer());
	  ArrowLineOvComp* comp = (ArrowLineOvComp*)ed->GetCameraComp();
	  ArrowLine* aline = comp->GetArrowLine();
	  aline->SetTransformer(rel);
	  //	      aline->SetOriginal(px0, py0, px1, py1);

	  ArrowLine* alinev = (ArrowLine*)(view->GetGraphic());
	  viewer->GetDamage()->Incur(alinev);
	  alinev->SetTransformer(rel);
	  alinev->SetOriginal(px0, py0, px1, py1);
	  viewer->GetDamage()->Incur(alinev);
	  unidraw->Update();
	}
	else {
	  Catalog* catalog = unidraw->GetCatalog();
	  ArrowLine* aline = new ArrowLine(px0, py0, px1, py1, false, true,
					   1.25, stdgraphic);
	  PSBrush* br = catalog->FindBrush(0xffff, 2);
	  aline->SetBrush(br);
	  PSColor* col = catalog->FindColor("Red");
	  aline->SetColors(col, col);
	  aline->SetTransformer(rel);
	  ArrowLineOvComp* comp = new ArrowLineOvComp(aline);
	  Command* pcmd = new PasteCmd(ed, new Clipboard(comp));
	  pcmd->Execute();
	  ed->GetSelection()->Clear();
	  ed->SetCameraComp(comp);
	}
    }
    return cmd;
}

/****************************************************************************/

#ifdef HAVE_ACE

UtmLosImportByServCmd::UtmLosImportByServCmd (ControlInfo* c) : UtmImportByServCmd(c) { }
UtmLosImportByServCmd::UtmLosImportByServCmd (Editor* ed) : UtmImportByServCmd(ed) { }

Command* UtmLosImportByServCmd::Copy () {
    Command* copy = new UtmLosImportByServCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void UtmLosImportByServCmd::LOSFunc(char* func) { _losfunc = func; }
void UtmLosImportByServCmd::DemPath(char* path) { _dempath = path; }
void UtmLosImportByServCmd::StartAzimuth(float az) { _startaz = az; }
void UtmLosImportByServCmd::EndAzimuth(float az) { _endaz = az; }
void UtmLosImportByServCmd::ObserverX(float x) { _obsx = x; }
void UtmLosImportByServCmd::ObserverY(float y) { _obsy = y; }
void UtmLosImportByServCmd::ObserverHeight(float h) { _obshgt = h; }
void UtmLosImportByServCmd::TargetX(float x) { _targx = x; }
void UtmLosImportByServCmd::TargetY(float y) { _targy = y; }
void UtmLosImportByServCmd::TargetHeight(float h) { _targhgt = h; }
void UtmLosImportByServCmd::TargetWidth(float w) { _targwid = w; }
void UtmLosImportByServCmd::Range(float r) { _range = r; }
void UtmLosImportByServCmd::ValueType(int v) { _valtype = v; }
void UtmLosImportByServCmd::LOSType(int v) { _lostype = v; }

void UtmLosImportByServCmd::Execute () {
  Catalog* catalog = unidraw->GetCatalog();
  const char* portstr = catalog->GetAttribute("losservport");
  u_short LOSSERV_PORT = atoi(portstr);
  const char *const LOSSERV_HOST = catalog->GetAttribute("losservhost");
  const char *losserv_host = LOSSERV_HOST;
  u_short losserv_port  = LOSSERV_PORT;
  ServHost(losserv_host);
  ServPort(losserv_port);

  char loscom[1024];
  sprintf(loscom, "painttbl(:reset);\
painttbl(\"vis==1||vis==2\" \":fgcolor \\\"red\\\",0,0,0 :brush 65535,0 :fillbg 0 :pattern 0x11,0x0,0x44,0x0,0x11,0x0,0x44,0x0\");\
painttbl(\"vis==3||vis==4\" \":fgcolor \\\"green\\\",0,0,0 :brush 65535,0 :fillbg 0 :pattern 0x11,0x0,0x44,0x0,0x11,0x0,0x44,0x0\");\
exportfc(%s(\"%s\" %f %f %f %f %f %f %f %f %f %f %d %d) \"localhost\" \"20011\" :paint )\n",
	  _losfunc,_dempath, _startaz, _endaz, _obsx, _obsy, _obshgt,
	  _targx, _targy, _targhgt, _targwid, _range, _valtype, _lostype);
  printf("los command: %s\n", loscom);
  ServCom(loscom);

#if 0
  UtmEditor* editor = (UtmEditor*)GetEditor();
  MapImportCmd* _import_cmd = new MapImportCmd(editor);
  ResultImportCmd(_import_cmd);
#endif
  UtmImportByServCmd::Execute();
}


/****************************************************************************/

ClassId UtmLosTool::GetClassId () { return UTMLOS_TOOL; }

boolean UtmLosTool::IsA (ClassId id) {
    return UTMLOS_TOOL == id || GraphicCompTool::IsA(id);
}

Tool* UtmLosTool::Copy () { return new UtmLosTool(CopyControlInfo()); }

UtmLosTool::UtmLosTool(ControlInfo* info)
:GraphicCompTool()
{
    if (info)
	SetControlInfo(info);
    ArrowLine* ml = new ArrowLine(
        0, 0, unit, unit, true, false, 2., stdgraphic
    );
    Catalog* catalog = unidraw->GetCatalog();
    PSBrush* br = catalog->FindBrush(0xffff, 2);
    ml->SetBrush(br);
    PSColor* col = catalog->FindColor("Red");
    ml->SetColors(col, col);
    ArrowLineOvComp* mlComp = new ArrowLineOvComp(ml);
    Init(mlComp);
}

Command* UtmLosTool::InterpretManipulator (Manipulator* m) {
    DragManip* dm = (DragManip*) m;
    MapEditor* ed = (MapEditor*) dm->GetViewer()->GetEditor();
    Transformer* rel = dm->GetTransformer();
    MacroCmd* cmd = new MacroCmd(ed);
    Transformer* trans = ed->GetMapViewState()->deftrans();

    RubberLine* rl = (RubberLine*) dm->GetRubberband();

    Coord px0, py0, px1, py1;
    int pt;
    rl->GetCurrent(px0, py0, px1, py1);

    if (rel != nil) {
      rel = new Transformer(rel);
      rel->Invert();
    }

    if (px0 != px1 || py0 != py1) {
        float fpx, fpy, ffx, ffy;
	rel->transform(px0, py0, fpx, fpy);
	rel->transform(px1, py1, ffx, ffy);

	//	printf("rl: %d %d %d %d; trl: %f %f %f %f\n",
	//	       px0, py0, px1, py1, fpx, fpy, ffx, ffy);

	if (trans) {
	  float tpx, tpy, tfx, tfy;
	  trans->inverse_transform(fpx, fpy, tpx, tpy);
	  trans->inverse_transform(ffx, ffy, tfx, tfy);

	  printf("UTM: P1: %fE %fN; P2: %fE %fN\n", tpx, tpy, tfx, tfy);

	  if (ed->GetCameraComp()) {
	    Viewer* viewer = ed->GetViewer();    
	    OverlayView* view =
	      ((OverlayComp*)ed->GetCameraComp())->FindView(ed->GetViewer());
	    ArrowLineOvComp* comp = (ArrowLineOvComp*)ed->GetCameraComp();
	    ArrowLine* aline = comp->GetArrowLine();
	    aline->SetTransformer(rel);
	    //	      aline->SetOriginal(px0, py0, px1, py1);
	    
	    ArrowLine* alinev = (ArrowLine*)(view->GetGraphic());
	    viewer->GetDamage()->Incur(alinev);
	    alinev->SetTransformer(rel);
	    alinev->SetOriginal(px0, py0, px1, py1);
	    viewer->GetDamage()->Incur(alinev);
	    unidraw->Update();
	  }
	  else {
	    Catalog* catalog = unidraw->GetCatalog();
	    ArrowLine* aline = new ArrowLine(px0, py0, px1, py1, false, true,
					     1.25, stdgraphic);
	    PSBrush* br = catalog->FindBrush(0xffff, 2);
	    aline->SetBrush(br);
	    PSColor* col = catalog->FindColor("Red");
	    aline->SetColors(col, col);
	    aline->SetTransformer(rel);
	    ArrowLineOvComp* comp = new ArrowLineOvComp(aline);
	    Command* pcmd = new PasteCmd(ed, new Clipboard(comp));
	    pcmd->Execute();
	    ed->GetSelection()->Clear();
	    ed->SetCameraComp(comp);
	    ed->GetViewer()->Update();
	  }

	  PJ* iproj = ed->GetMapViewState()->implicitproj();
	  if (iproj) {
	    UV data;
	    data.u = tpx;
	    data.v = tpy;
	    data = pj_inv(data, iproj);
	    float obsx = data.u * RAD_TO_DEG;
	    float obsy = data.v * RAD_TO_DEG;
	    data.u = tfx;
	    data.v = tfy;
	    data = pj_inv(data, iproj);
	    float tarx = data.u * RAD_TO_DEG;
	    float tary = data.v * RAD_TO_DEG;
	    float dist = sqrt((tfy - tpy)*(tfy - tpy) + 
			      (tfx - tpx)*(tfx - tpx));
	    UtmLosImportByServCmd* loscmd = new UtmLosImportByServCmd(ed);
	    Catalog* catalog = unidraw->GetCatalog();
	    const char* dempath = catalog->GetAttribute("dem");
	    if (strcmp(dempath, "") != 0) {
	      loscmd->LOSFunc("demlos");
	      loscmd->DemPath((char*)dempath);
	    }
	    else {
	      const char* dtedpath = catalog->GetAttribute("dtedlos");
	      if (strcmp(dtedpath, "") != 0) {
		loscmd->LOSFunc("dtedlos");
		loscmd->DemPath((char*)dtedpath);
	      }
	    }
	    loscmd->StartAzimuth(0.0);
	    loscmd->EndAzimuth(360.0);
	    loscmd->ObserverX(obsx);
	    loscmd->ObserverY(obsy);
	    loscmd->ObserverHeight(2.0);
	    loscmd->TargetX(tarx);
	    loscmd->TargetY(tary);
	    loscmd->TargetHeight(2.0);
	    loscmd->TargetWidth(2.0);
	    loscmd->Range(dist + 50.0);
	    loscmd->ValueType(LOS_4VALUED);
	    loscmd->LOSType(LOS_VIEWSHED);
	    //cmd->Append(loscmd);
	    loscmd->Execute();
	  }
	}
    }
    return cmd;
}


/****************************************************************************/

UtmPlayRouteCmd::UtmPlayRouteCmd(ControlInfo* ci): Command(ci) {
}

Command* UtmPlayRouteCmd::Copy () {
    Command* copy = new UtmPlayRouteCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void UtmPlayRouteCmd::Execute() {
  Selection* s = GetEditor()->GetViewer()->GetSelection();
  if (s->Number() > 0) {
    Iterator i;
    s->First(i);
    GraphicView* view = s->GetView(i);
    if (view->IsA(MAP_ROUTE_VIEW)) {
      ACE_SOCK_Stream* _socket = new ACE_SOCK_Stream;
      ACE_SOCK_Connector* connector = new ACE_SOCK_Connector;
      ACE_INET_Addr* addr = new ACE_INET_Addr(20017, "localhost");

      if (connector->connect (*_socket, *addr) == -1) {
	GetEditor()->GetWindow()->cursor(arrow);
	ACE_ERROR ((LM_ERROR, "%p\n", "open"));
	return;
      }

      MapRouteComp* rtcomp = (MapRouteComp*)((OverlayView*)view)->GetOverlayComp();
      AttributeList* al = rtcomp->GetAttributeList();
      if (!al->find("name")) {
	al->add_attr("name",
		     new AttributeValue(symbol_add("route_utmviewer"),
					AttributeValue::StringType));
      }

      char com1[BUFSIZ];
      MapRouteScript* rtscript = (MapRouteScript*)rtcomp->Create(SCRIPT_VIEW);
      rtcomp->Attach(rtscript);
      strstreambuf sbuf(com1, BUFSIZ);
      ostream ostr(&sbuf);
      rtscript->Definition(ostr);
      ostr.put('\0');
      ostr.flush();
      int len = strlen(com1);
      for (int j=0; j<len; j++)  // kludge because multiple lines not 
	if (com1[j]=='\n') com1[j] = ' '; // yet supported by ACE comterp

      char com2[] = "create(vehicle_utmviewer);connect(vehicle_utmviewer \"localhost\" 20012);connect(vehicle_utmviewer \"localhost\" 20010);playroute(%s vehicle_utmviewer :speed 30)\n";

      filebuf fbuf;
      fbuf.attach(_socket->get_handle());
      ostream out(&fbuf);
      out << com1 << ";";
      out.form( com2, al->find("name")->string_ptr());
    }
  }
}

#endif /* defined(HAVE_ACE) */

/*****************************************************************************/

ClassId UtmOpenCmd::GetClassId () { return UTM_OPEN_CMD; }

boolean UtmOpenCmd::IsA (ClassId id) {
    return UTM_OPEN_CMD == id || MapOpenCmd::IsA(id);
}

UtmOpenCmd::UtmOpenCmd (ControlInfo* c, OpenFileChooser* fc) : MapOpenCmd(c, fc) { }
UtmOpenCmd::UtmOpenCmd (Editor* ed, OpenFileChooser* fc) : MapOpenCmd(ed, fc) { }

Command* UtmOpenCmd::Copy () {
    Command* copy = new UtmOpenCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void UtmOpenCmd::Execute () {
    UtmEditor* ed = (UtmEditor*)GetEditor();
    MapIdrawView::SetMapEditor(ed);
    OvOpenCmd::Execute();
    ed->AdjustForGeo();
}

#ifdef HAVE_ACE

/****************************************************************************/

UtmElevImportCmd::UtmElevImportCmd (ControlInfo* c) : UtmImportByServCmd(c) { }
UtmElevImportCmd::UtmElevImportCmd (Editor* ed) : UtmImportByServCmd(ed) { }

Command* UtmElevImportCmd::Copy () {
    Command* copy = new UtmElevImportCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void UtmElevImportCmd::ElevFunc(char* f) { _elevfunc = f; }
void UtmElevImportCmd::DemPath(char* path) { _dempath = path; }
void UtmElevImportCmd::ObserverX(float x) { _obsx = x; }
void UtmElevImportCmd::ObserverY(float y) { _obsy = y; }

void UtmElevImportCmd::Execute () {
  Catalog* catalog = unidraw->GetCatalog();
  const char* portstr = catalog->GetAttribute("losservport");
  u_short LOSSERV_PORT = atoi(portstr);
  const char *const LOSSERV_HOST = catalog->GetAttribute("losservhost");
  const char *losserv_host = LOSSERV_HOST;
  u_short losserv_port  = LOSSERV_PORT;
  ServHost(losserv_host);
  ServPort(losserv_port);

  char loscom[1024];
  sprintf(loscom, "painttbl(:reset);\
painttbl(\"1==1\" \":fgcolor \\\"purple\\\",0,0,0 :brush 65535,0 :fillbg 0 :pattern 0x11,0x0,0x44,0x0,0x11,0x0,0x44,0x0\");\
exportfc(%s(\"%s\" %f %f) \"localhost\" \"20011\" :paint )\n",
	  _elevfunc, _dempath, _obsx, _obsy);
  printf("elev command: %s\n", loscom);
  ServCom(loscom);

  UtmImportByServCmd::Execute();
}

/****************************************************************************/

ClassId UtmElevTool::GetClassId () { return UTMELEV_TOOL; }

boolean UtmElevTool::IsA (ClassId id) {
    return UTMELEV_TOOL == id || GraphicCompTool::IsA(id);
}

Tool* UtmElevTool::Copy () { return new UtmElevTool(CopyControlInfo()); }

UtmElevTool::UtmElevTool(ControlInfo* info)
:GraphicCompTool()
{
    if (info)
	SetControlInfo(info);
    ArrowLine* ml = new ArrowLine(
        0, 0, unit, unit, true, false, 2., stdgraphic
    );
    Catalog* catalog = unidraw->GetCatalog();
    PSBrush* br = catalog->FindBrush(0xffff, 2);
    ml->SetBrush(br);
    PSColor* col = catalog->FindColor("Red");
    ml->SetColors(col, col);
    ArrowLineOvComp* mlComp = new ArrowLineOvComp(ml);
    Init(mlComp);
}

Command* UtmElevTool::InterpretManipulator (Manipulator* m) {
    DragManip* dm = (DragManip*) m;
    MapEditor* ed = (MapEditor*) dm->GetViewer()->GetEditor();
    Transformer* rel = dm->GetTransformer();
    MacroCmd* cmd = new MacroCmd(ed);
    Transformer* trans = ed->GetMapViewState()->deftrans();

    RubberLine* rl = (RubberLine*) dm->GetRubberband();

    Coord px0, py0, px1, py1;
    int pt;
    rl->GetCurrent(px0, py0, px1, py1);

    if (rel != nil) {
      rel = new Transformer(rel);
      rel->Invert();
    }

    if (px0 != px1 || py0 != py1) {
        float fpx, fpy, ffx, ffy;
	rel->transform(px0, py0, fpx, fpy);
	rel->transform(px1, py1, ffx, ffy);

	//	printf("rl: %d %d %d %d; trl: %f %f %f %f\n",
	//	       px0, py0, px1, py1, fpx, fpy, ffx, ffy);

	if (trans) {
	  float tpx, tpy, tfx, tfy;
	  trans->inverse_transform(fpx, fpy, tpx, tpy);
	  trans->inverse_transform(ffx, ffy, tfx, tfy);

	  printf("UTM: P1: %fE %fN; P2: %fE %fN\n", tpx, tpy, tfx, tfy);

	  if (ed->GetCameraComp()) {
	    Viewer* viewer = ed->GetViewer();    
	    OverlayView* view =
	      ((OverlayComp*)ed->GetCameraComp())->FindView(ed->GetViewer());
	    ArrowLineOvComp* comp = (ArrowLineOvComp*)ed->GetCameraComp();
	    ArrowLine* aline = comp->GetArrowLine();
	    aline->SetTransformer(rel);
	    //	      aline->SetOriginal(px0, py0, px1, py1);
	    
	    ArrowLine* alinev = (ArrowLine*)(view->GetGraphic());
	    viewer->GetDamage()->Incur(alinev);
	    alinev->SetTransformer(rel);
	    alinev->SetOriginal(px0, py0, px1, py1);
	    viewer->GetDamage()->Incur(alinev);
	    unidraw->Update();
	  }
	  else {
	    Catalog* catalog = unidraw->GetCatalog();
	    ArrowLine* aline = new ArrowLine(px0, py0, px1, py1, false, true,
					     1.25, stdgraphic);
	    PSBrush* br = catalog->FindBrush(0xffff, 2);
	    aline->SetBrush(br);
	    PSColor* col = catalog->FindColor("Red");
	    aline->SetColors(col, col);
	    aline->SetTransformer(rel);
	    ArrowLineOvComp* comp = new ArrowLineOvComp(aline);
	    Command* pcmd = new PasteCmd(ed, new Clipboard(comp));
	    pcmd->Execute();
	    ed->GetSelection()->Clear();
	    ed->SetCameraComp(comp);
	    ed->GetViewer()->Update();
	  }

	  PJ* iproj = ed->GetMapViewState()->implicitproj();
	  if (iproj) {
	    UV data;
	    data.u = tpx;
	    data.v = tpy;
	    data = pj_inv(data, iproj);
	    float obsx = data.u * RAD_TO_DEG;
	    float obsy = data.v * RAD_TO_DEG;
	    data.u = tfx;
	    data.v = tfy;
	    data = pj_inv(data, iproj);
	    float tarx = data.u * RAD_TO_DEG;
	    float tary = data.v * RAD_TO_DEG;
	    float dist = sqrt((tfy - tpy)*(tfy - tpy) + 
			      (tfx - tpx)*(tfx - tpx));
	    UtmElevImportCmd* elevcmd = new UtmElevImportCmd(ed);
	    Catalog* catalog = unidraw->GetCatalog();
	    const char* dempath = catalog->GetAttribute("dem");
	    if (strcmp(dempath, "") != 0) {
	      elevcmd->ElevFunc("demtriutm");
	      elevcmd->DemPath((char*)dempath);
	    }
	    else {
	      const char* dtedpath = catalog->GetAttribute("dtedlos");
	      if (strcmp(dtedpath, "") != 0) {
		elevcmd->ElevFunc("dtedtriutm");
		elevcmd->DemPath((char*)dtedpath);
	      }
	    }
	    elevcmd->ObserverX(tpx);
	    elevcmd->ObserverY(tpy);
	    elevcmd->Execute();
	  }
	}
    }
    return cmd;
}

/****************************************************************************/

UtmElevGridImportCmd::UtmElevGridImportCmd (ControlInfo* c) : UtmImportByServCmd(c) { }
UtmElevGridImportCmd::UtmElevGridImportCmd (Editor* ed) : UtmImportByServCmd(ed) { }

Command* UtmElevGridImportCmd::Copy () {
    Command* copy = new UtmElevGridImportCmd(CopyControlInfo());
    InitCopy(copy);
    return copy;
}

void UtmElevGridImportCmd::ElevFunc(char* f) { _elevfunc = f; }
void UtmElevGridImportCmd::DemPath(char* path) { _dempath = path; }
void UtmElevGridImportCmd::Corners(float x0, float y0, float x1, float y1) {
  _x0 = x0;
  _y0 = y0;
  _x1 = x1;
  _y1 = y1;
}

void UtmElevGridImportCmd::Execute () {
  Catalog* catalog = unidraw->GetCatalog();
  const char* portstr = catalog->GetAttribute("losservport");
  u_short LOSSERV_PORT = atoi(portstr);
  const char *const LOSSERV_HOST = catalog->GetAttribute("losservhost");
  const char *losserv_host = LOSSERV_HOST;
  u_short losserv_port  = LOSSERV_PORT;
  ServHost(losserv_host);
  ServPort(losserv_port);

  char loscom[1024];
  sprintf(loscom, "painttbl(:reset);\
painttbl(\"1==1\" \":fgcolor \\\"purple\\\",0,0,0 :brush 65535,0 :fillbg 0 :pattern 0x11,0x0,0x44,0x0,0x11,0x0,0x44,0x0\");\
exportfc(%s(\"%s\" %f %f %f %f) \"localhost\" \"20011\" :paint )\n",
	  _elevfunc, _dempath, _x0, _y0, _x1, _y1);
  printf("elev command: %s\n", loscom);
  ServCom(loscom);

  UtmImportByServCmd::Execute();
}

/****************************************************************************/

ClassId UtmElevGridTool::GetClassId () { return UTMELEVGRID_TOOL; }

boolean UtmElevGridTool::IsA (ClassId id) {
    return UTMELEVGRID_TOOL == id || GraphicCompTool::IsA(id);
}

Tool* UtmElevGridTool::Copy () { return new UtmElevGridTool(CopyControlInfo()); }

UtmElevGridTool::UtmElevGridTool(ControlInfo* info)
:GraphicCompTool()
{
    if (info)
	SetControlInfo(info);
    SF_Rect* ml = new SF_Rect(
        0, 0, unit, unit, stdgraphic
    );
    Catalog* catalog = unidraw->GetCatalog();
    PSBrush* br = catalog->FindBrush(0xffff, 2);
    ml->SetBrush(br);
    PSColor* col = catalog->FindColor("Red");
    ml->SetColors(col, col);
    ml->SetPattern(psnonepat);
    RectOvComp* mlComp = new RectOvComp(ml);
    Init(mlComp);
}

Command* UtmElevGridTool::InterpretManipulator (Manipulator* m) {
    DragManip* dm = (DragManip*) m;
    MapEditor* ed = (MapEditor*) dm->GetViewer()->GetEditor();
    Transformer* rel = dm->GetTransformer();
    MacroCmd* cmd = new MacroCmd(ed);
    Transformer* trans = ed->GetMapViewState()->deftrans();

    RubberRect* rl = (RubberRect*) dm->GetRubberband();

    Coord px0, py0, px1, py1;
    int pt;
    rl->GetCurrent(px0, py0, px1, py1);
    Coord mpx0 = Math::min(px0, px1);
    Coord mpy0 = Math::min(py0, py1);
    Coord mpx1 = Math::max(px0, px1);
    Coord mpy1 = Math::max(py0, py1);

    if (rel != nil) {
      rel = new Transformer(rel);
      rel->Invert();
    }

    if (mpx0 != mpx1 || mpy0 != mpy1) {
        float fpx, fpy, ffx, ffy;
	rel->transform(mpx0, mpy0, fpx, fpy);
	rel->transform(mpx1, mpy1, ffx, ffy);

	//	printf("rl: %d %d %d %d; trl: %f %f %f %f\n",
	//	       px0, py0, px1, py1, fpx, fpy, ffx, ffy);

	if (trans) {
	  float tpx, tpy, tfx, tfy;
	  trans->inverse_transform(fpx, fpy, tpx, tpy);
	  trans->inverse_transform(ffx, ffy, tfx, tfy);

	  printf("UTM: P1: %fE %fN; P2: %fE %fN\n", tpx, tpy, tfx, tfy);

#if 0
	  if (ed->GetCameraComp()) {
	    Viewer* viewer = ed->GetViewer();    
	    OverlayView* view =
	      ((OverlayComp*)ed->GetCameraComp())->FindView(ed->GetViewer());
	    ArrowLineOvComp* comp = (ArrowLineOvComp*)ed->GetCameraComp();
	    ArrowLine* aline = comp->GetArrowLine();
	    aline->SetTransformer(rel);
	    //	      aline->SetOriginal(px0, py0, px1, py1);
	    
	    ArrowLine* alinev = (ArrowLine*)(view->GetGraphic());
	    viewer->GetDamage()->Incur(alinev);
	    alinev->SetTransformer(rel);
	    alinev->SetOriginal(px0, py0, px1, py1);
	    viewer->GetDamage()->Incur(alinev);
	    unidraw->Update();
	  }
	  else {
#endif
	    Catalog* catalog = unidraw->GetCatalog();
	    SF_Rect* aline = new SF_Rect(mpx0, mpy0, mpx1, mpy1, stdgraphic);
	    PSBrush* br = catalog->FindBrush(0xffff, 2);
	    aline->SetBrush(br);
	    PSColor* col = catalog->FindColor("Red");
	    aline->SetColors(col, col);
	    aline->SetPattern(psnonepat);
	    aline->SetTransformer(rel);
	    RectOvComp* comp = new RectOvComp(aline);
	    Command* pcmd = new PasteCmd(ed, new Clipboard(comp));
	    pcmd->Execute();
	    ed->GetSelection()->Clear();
	    //ed->SetCameraComp(comp);
	    ed->GetViewer()->Update();
#if 0
	  }
#endif

	  PJ* iproj = ed->GetMapViewState()->implicitproj();
	  if (iproj) {
	    UV data;
	    data.u = tpx;
	    data.v = tpy;
	    data = pj_inv(data, iproj);
	    float obsx = data.u * RAD_TO_DEG;
	    float obsy = data.v * RAD_TO_DEG;
	    data.u = tfx;
	    data.v = tfy;
	    data = pj_inv(data, iproj);
	    float tarx = data.u * RAD_TO_DEG;
	    float tary = data.v * RAD_TO_DEG;

	    UtmElevGridImportCmd* elevcmd = new UtmElevGridImportCmd(ed);
	    Catalog* catalog = unidraw->GetCatalog();
	    const char* dempath = catalog->GetAttribute("dem");
	    if (strcmp(dempath, "") != 0) {
	      elevcmd->ElevFunc("demgridutm");
	      elevcmd->DemPath((char*)dempath);
	    }
	    else {
	      const char* dtedpath = catalog->GetAttribute("dtedlos");
	      if (strcmp(dtedpath, "") != 0) {
		elevcmd->ElevFunc("dtedgridutm");
		elevcmd->DemPath((char*)dtedpath);
	      }
	    }
	    elevcmd->Corners(tpx, tpy, tfx, tfy);
	    elevcmd->Execute();
	  }
	}
    }
    return cmd;
}

#endif /* defined(HAVE_ACE) */
